    MODULE cdf_normal_mod
!----------------------------------------------------------------------
!      SUBROUTINE CDF_NORMAL( WHICH, CUM, CCUM, X, MEAN, SD, &
!                             STATUS, BOUND, CHECK_INPUT )
!                   Cumulative Distribution Function
!                         NORMAL Distribution
!        DENSITY OF THE DISTRIBUTION (IN x) IS PROPORTIONAL TO
!                     exp( -(x-mean)**2 / 2*sd**2
!                              FUNCTION
!     Calculates any one parameter of the normal distribution given
!     values for the others.
!                                 NOTE
!     dpkind is double precision
!                              ARGUMENTS
!     WHICH --> Integer indicating which of the next four argument
!               values is to be calculated from the others.
!               Legal range: 1..4
!               iwhich = 1 : Calculate CUM and CCUM from X,MEAN, and SD
!               iwhich = 2 : Calculate X from CUM,CCUM,MEAN, and SD
!               iwhich = 3 : Calculate MEAN from CUM,CCUM,X, and SD
!               iwhich = 4 : Calculate SD from CUM,CCUM,X, and MEAN
!                    INTEGER, INTENT(IN):: WHICH
!     CUM <--> The integral from -infinity to X of the normal
!            distribution.
!            Input range: [1e-8, 1-1e-8].
!     REAL (dpkind), OPTIONAL :: CUM
!     CCUM <--> 1-CUM.
!            Input range: [1e-8, 1-1e-8].
!            CUM + CCUM = 1
!            REAL (dpkind), OPTIONAL :: CCUM
! NOTE: At least one argument CUM or CCUM must be present and both may
!       be present.  If both arguments are input then their values must
!       sum to one.
!     X <--> Upper limit of integration of the normal density.
!            Input range: Unlimited
!            Search range: Unlimited
!             REAL (dpkind) :: X
!     MEAN <--> The mean of the normal distribution
!            Input range: Unlimited
!            Search range: Unlimited
!            REAL (dpkind), OPTIONAL :: MEAN
! NOTE: If the argument MEAN is not present, the mean is set to
!       zero.
!     SD <--> The standard deviation of the normal distribution
!            Input range: (1E-10, 1E100).
!            Search range: Unlimited
!            REAL (dpkind), OPTIONAL :: SD
! NOTE: If the argument SD is not present, the sd is set to one.
!     STATUS <-- The status of the calculation.  Zero if it completed
!                correctly.  See table below.
!                INTEGER, OPTIONAL, INTENT(OUT) :: STATUS
! NOTE: If STATUS is not present in the call and a status other than zero
! would have been returned, the program aborts with an error message.
!     BOUND <-- The bound exceeded by some item (see STATUS) causing
!               an error return.
!                REAL (dpkind), OPTIONAL, INTENT(OUT) :: BOUND
!     CHECK_INPUT --> .TRUE. if input values are to be checked for
!                     legality.  If .FALSE., the values are not checked
!                     and the programmer of the code using this routine
!                     has reponsibility for their legality.
!                     LOGICAL, OPTIONAL, INTENT(IN) :: CHECK_INPUT
! NOTE: If CHECK_INPUT is not present in the call, it takes the value
! .TRUE.
!----------
!          Parameters are numbered as follows for use with STATUS:
!                1 - which
!                2 - cum
!                3 - ccum
!                4 - x
!                5 - mean
!                6 - sd
!----------
!     STATUS <-- 0 if calculation completed correctly
!               -I if input parameter number I is out of range
!                3 if CUM + CCUM .ne. 1
!----------------------------------------------------------------------
!----------------------------------------------------------------------
!      FUNCTION CUM_NORMAL(X,MEAN,SD,STATUS,BOUND,CHECK_INPUT)
!        REAL (dpkind) :: CUM_NORMAL
!      Returns CUM from cdf_normal.  All arguments are as per
!      cdf_normal.
!----------------------------------------------------------------------
!----------------------------------------------------------------------
!      FUNCTION CCUM_NORMAL(X,MEAN,SD,STATUS,BOUND,CHECK_INPUT)
!        REAL (dpkind) :: CUM_NORMAL
!      Returns CCUM from cdf_normal.  All arguments are as per
!      cdf_normal.
!----------------------------------------------------------------------
!----------------------------------------------------------------------
!      FUNCTION inv_normal(cum,ccum,mean,sd,status,bound,check_input)
!        REAL (dpkind) :: inv_normal
!      Returns X from cdf_normal.  All arguments are as per
!      cdf_normal
!----------------------------------------------------------------------
!----------------------------------------------------------------------
! .. Use Statements ..
      USE cdf_aux_mod
! ..
! .. Implicit None Statement ..
      IMPLICIT NONE
! ..
! .. Default Accessibility ..
      PRIVATE
! ..
! .. Parameters ..
      INTEGER, PARAMETER :: dpkind = KIND(0.0D0)
      REAL (dpkind), PARAMETER :: half = 0.5_dpkind
      REAL (dpkind), PARAMETER :: one = 1.0_dpkind
      REAL (dpkind), PARAMETER :: zero = 0.0_dpkind
! ..
! .. Public Statements ..
      PUBLIC :: ccum_normal, cdf_normal, cum_normal, inv_normal
! ..
    CONTAINS

!*********************************************************************

      FUNCTION ccum_normal(x,mean,sd,status,bound,check_input)
! .. Implicit None Statement ..
        IMPLICIT NONE
! ..
! .. Function Return Value ..
        REAL (dpkind) :: ccum_normal
! ..
! .. Scalar Arguments ..
        REAL (dpkind), OPTIONAL :: bound, mean, sd
        REAL (dpkind) :: x
        INTEGER, OPTIONAL, INTENT (OUT) :: status
        LOGICAL, OPTIONAL, INTENT (IN) :: check_input
! ..
! .. Executable Statements ..

        CALL cdf_normal(which=1,ccum=ccum_normal,x=x,mean=mean,sd=sd, &
          status=status,bound=bound,check_input=check_input)

      END FUNCTION ccum_normal

!*********************************************************************

      SUBROUTINE cdf_normal(which,cum,ccum,x,mean,sd,status,bound, &
          check_input)
! .. Implicit None Statement ..
        IMPLICIT NONE
! ..
! .. Scalar Arguments ..
        REAL (dpkind), OPTIONAL :: bound, ccum, cum, mean, sd
        REAL (dpkind) :: x
        INTEGER, OPTIONAL, INTENT (OUT) :: status
        INTEGER, INTENT (IN) :: which
        LOGICAL, OPTIONAL, INTENT (IN) :: check_input
! ..
! .. Local Arrays ..
        REAL (dpkind) :: params(6)
! ..
! .. Local Scalars ..
        REAL (dpkind) :: local_ccum, local_cum, local_mean, local_sd, z
        LOGICAL :: local_check_input
! ..
! .. Intrinsic Functions ..
        INTRINSIC PRESENT
! ..
! .. Executable Statements ..

        IF (PRESENT(status)) status = 0

! Check presence of arguments and fix up


        CALL check_complements(cum,ccum,the_normal%name,'cum','ccum', &
          local_cum,local_ccum,set_values=which/=1)

        IF (PRESENT(mean)) THEN
          local_mean = mean
        ELSE
          local_mean = zero
        END IF

        IF (PRESENT(sd)) THEN
          local_sd = sd
        ELSE
          local_sd = one
        END IF

        params(1) = local_cum
        params(2) = local_ccum
        params(3) = x
        params(4) = local_mean
        params(5) = local_sd

        IF (PRESENT(check_input)) THEN
          local_check_input = check_input
        ELSE
          local_check_input = .TRUE.
        END IF

!!! Check for consistency of which and answer

        IF (local_check_input) THEN

          IF (which==3 .AND. .NOT. PRESENT(mean)) CALL which_miss(which, &
            the_normal%name,'mean')

          IF (which==4 .AND. .NOT. PRESENT(sd)) CALL which_miss(which, &
            the_normal%name,'sd')

          CALL validateparameters(the_normal,which,params,status,bound)
        END IF

! ++++++++++          ++++++++++          ++++++++++
! Compute the Answers
! ++++++++++          ++++++++++          ++++++++++

        SELECT CASE (which)
        CASE (1)
          z = (x-local_mean)/local_sd

          CALL local_cum_normal(z,local_cum,local_ccum)

          IF (PRESENT(cum)) cum = local_cum
          IF (PRESENT(ccum)) ccum = local_ccum

        CASE (2)
          z = local_inv_normal(local_cum,local_ccum)

          x = local_sd*z + local_mean

        CASE (3)
          z = local_inv_normal(local_cum,local_ccum)

          mean = x - local_sd*z

        CASE (4)
          z = local_inv_normal(local_cum,local_ccum)

          sd = (x-local_mean)/z
        END SELECT

      END SUBROUTINE cdf_normal

!*********************************************************************

      FUNCTION cum_normal(x,mean,sd,status,bound,check_input)
! .. Implicit None Statement ..
        IMPLICIT NONE
! ..
! .. Function Return Value ..
        REAL (dpkind) :: cum_normal
! ..
! .. Scalar Arguments ..
        REAL (dpkind), OPTIONAL :: bound, mean, sd
        REAL (dpkind) :: x
        INTEGER, OPTIONAL, INTENT (OUT) :: status
        LOGICAL, OPTIONAL, INTENT (IN) :: check_input
! ..
! .. Executable Statements ..

        CALL cdf_normal(which=1,cum=cum_normal,x=x,mean=mean,sd=sd, &
          status=status,bound=bound,check_input=check_input)

      END FUNCTION cum_normal

!*********************************************************************

      FUNCTION inv_normal(cum,ccum,mean,sd,status,bound,check_input)
! .. Implicit None Statement ..
        IMPLICIT NONE
! ..
! .. Function Return Value ..
        REAL (dpkind) :: inv_normal
! ..
! .. Scalar Arguments ..
        REAL (dpkind), OPTIONAL :: bound, ccum, cum, mean, sd
        INTEGER, OPTIONAL, INTENT (OUT) :: status
        LOGICAL, OPTIONAL, INTENT (IN) :: check_input
! ..
! .. Executable Statements ..

        CALL cdf_normal(which=2,cum=cum,ccum=ccum,x=inv_normal,mean=mean, &
          sd=sd,status=status,bound=bound,check_input=check_input)

      END FUNCTION inv_normal

!*********************************************************************

      SUBROUTINE local_cum_normal(arg,result,ccum)
!----------------------------------------------------------------------
!     SUBROUINE LOCAL_CUM_NORMAL(X,RESULT,CCUM)
!                              Function
!     Computes the cumulative  of    the  normal   distribution,   i.e.,
!     the integral from -infinity to x of
!          (1/sqrt(2*pi)) exp(-u*u/2) du
!     X --> Upper limit of integration.
!                                        X is DOUBLE PRECISION
!     RESULT <-- Cumulative normal distribution.
!                                        RESULT is DOUBLE PRECISION
!     CCUM <-- Compliment of Cumulative normal distribution.
!                                        CCUM is DOUBLE PRECISION
!     Renaming of function ANORM from:
!     Cody, W.D. (1993). "ALGORITHM 715: SPECFUN - A Portabel FORTRAN
!     Package of Special Function Routines and Test Drivers"
!     acm Transactions on Mathematical Software. 19, 22-32.
!     with slight modifications to return ccum and to deal with
!     machine constants.
!----------------------------------------------------------------------
! Original Comments:
!------------------------------------------------------------------
! This function evaluates the normal distribution function:
!                              / x
!                     1       |       -t*t/2
!          P(x) = ----------- |      e       dt
!                 sqrt(2 pi)  |
!                             /-oo
!   The main computation evaluates near-minimax approximations
!   derived from those in "Rational Chebyshev approximations for
!   the error function" by W. J. Cody, Math. Comp., 1969, 631-637.
!   This transportable program uses rational functions that
!   theoretically approximate the normal distribution function to
!   at least 18 significant decimal digits.  The accuracy achieved
!   depends on the arithmetic system, the compiler, the intrinsic
!   functions, and proper selection of the machine-dependent
!   constants.
!*******************************************************************
!*******************************************************************
! Explanation of machine-dependent constants.
!   MIN   = smallest machine representable number.
!   EPS   = argument below which anorm(x) may be represented by
!           0.5  and above which  x*x  will not underflow.
!           A conservative value is the largest machine number X
!           such that   1.0 + X = 1.0   to machine precision.
!*******************************************************************
!*******************************************************************
! Error returns
!  The program returns  ANORM = 0     for  ARG .LE. XLOW.
! Intrinsic functions required are:
!     ABS, AINT, EXP
!  Author: W. J. Cody
!          Mathematics and Computer Science Division
!          Argonne National Laboratory
!          Argonne, IL 60439
!  Latest modification: March 15, 1992
!------------------------------------------------------------------
!------------------------------------------------------------------
!  Mathematical constants
!  SQRPI = 1 / sqrt(2*pi), ROOT32 = sqrt(32), and
!  THRSH is the argument for which anorm = 0.75.
!------------------------------------------------------------------
!------------------------------------------------------------------
!  Coefficients for approximation in first interval
!------------------------------------------------------------------
!------------------------------------------------------------------
!  Coefficients for approximation in second interval
!------------------------------------------------------------------
!------------------------------------------------------------------
!  Coefficients for approximation in third interval
!------------------------------------------------------------------
! .. Implicit None Statement ..
        IMPLICIT NONE
! ..
! .. Scalar Arguments ..
        REAL (dpkind) :: arg, ccum, result
! ..
! .. Local Scalars ..
        REAL (dpkind) :: del, eps, min, temp, x, xden, xnum, xsq, y
        INTEGER :: i
! ..
! .. Intrinsic Functions ..
        INTRINSIC ABS, AINT, EPSILON, EXP, TINY
! ..
! .. Parameters ..
        REAL (dpkind), PARAMETER :: a(5) = (/ &
          2.2352520354606839287E00_dpkind, &
          1.6102823106855587881E02_dpkind, &
          1.0676894854603709582E03_dpkind, &
          1.8154981253343561249E04_dpkind, &
          6.5682337918207449113E-2_dpkind/)
        REAL (dpkind), PARAMETER :: b(4) = (/ &
          4.7202581904688241870E01_dpkind, &
          9.7609855173777669322E02_dpkind, &
          1.0260932208618978205E04_dpkind, &
          4.5507789335026729956E04_dpkind/)
        REAL (dpkind), PARAMETER :: c(9) = (/ &
          3.9894151208813466764E-1_dpkind, &
          8.8831497943883759412E00_dpkind, &
          9.3506656132177855979E01_dpkind, &
          5.9727027639480026226E02_dpkind, &
          2.4945375852903726711E03_dpkind, &
          6.8481904505362823326E03_dpkind, &
          1.1602651437647350124E04_dpkind, &
          9.8427148383839780218E03_dpkind, &
          1.0765576773720192317E-8_dpkind/)
        REAL (dpkind), PARAMETER :: d(8) = (/ &
          2.2266688044328115691E01_dpkind, &
          2.3538790178262499861E02_dpkind, &
          1.5193775994075548050E03_dpkind, &
          6.4855582982667607550E03_dpkind, &
          1.8615571640885098091E04_dpkind, &
          3.4900952721145977266E04_dpkind, &
          3.8912003286093271411E04_dpkind, &
          1.9685429676859990727E04_dpkind/)
        REAL (dpkind), PARAMETER :: p(6) = (/ &
          2.1589853405795699E-1_dpkind, 1.274011611602473639E-1_dpkind, &
          2.2235277870649807E-2_dpkind, 1.421619193227893466E-3_dpkind, &
          2.9112874951168792E-5_dpkind, 2.307344176494017303E-2_dpkind/)
        REAL (dpkind), PARAMETER :: q(5) = (/ &
          1.28426009614491121E00_dpkind, 4.68238212480865118E-1_dpkind, &
          6.59881378689285515E-2_dpkind, 3.78239633202758244E-3_dpkind, &
          7.29751555083966205E-5_dpkind/)
        REAL (dpkind), PARAMETER :: root32 = 5.656854248E0_dpkind
        REAL (dpkind), PARAMETER :: sixten = 1.6_dpkind
        REAL (dpkind), PARAMETER :: sqrpi = &
          3.9894228040143267794E-1_dpkind
        REAL (dpkind), PARAMETER :: thrsh = 0.66291E0_dpkind
! ..
! .. Executable Statements ..

!------------------------------------------------------------------
!  Machine dependent constants
!------------------------------------------------------------------
!        eps = spmpar(1)*0.5E0_dpkind
        eps = EPSILON(one)*half
!        min = spmpar(2)
        min = TINY(one)
!------------------------------------------------------------------
        x = arg
        y = ABS(x)
        IF (y<=thrsh) THEN
!------------------------------------------------------------------
!  Evaluate  anorm  for  |X| <= 0.66291
!------------------------------------------------------------------
          xsq = zero
          IF (y>eps) xsq = x*x
          xnum = a(5)*xsq
          xden = xsq
          xnum = (xnum+a(1))*xsq
          xden = (xden+b(1))*xsq
          xnum = (xnum+a(2))*xsq
          xden = (xden+b(2))*xsq
          xnum = (xnum+a(3))*xsq
          xden = (xden+b(3))*xsq
          result = x*(xnum+a(4))/(xden+b(4))
          temp = result
          result = half + temp
          ccum = half - temp
!------------------------------------------------------------------
!  Evaluate  anorm  for 0.66291 <= |X| <= sqrt(32)
!------------------------------------------------------------------
        ELSE IF (y<=root32) THEN
          xnum = c(9)*y
          xden = y

          DO i = 1, 7
            xnum = (xnum+c(i))*y
            xden = (xden+d(i))*y
          END DO

          result = (xnum+c(8))/(xden+d(8))
          xsq = AINT(y*sixten)/sixten
          del = (y-xsq)*(y+xsq)
          result = EXP(-xsq*xsq*half)*EXP(-del*half)*result
          ccum = one - result

          IF (x>zero) THEN
            temp = result
            result = ccum
            ccum = temp
          END IF

!------------------------------------------------------------------
!  Evaluate  anorm  for |X| > sqrt(32)
!------------------------------------------------------------------
        ELSE
          result = zero
          xsq = one/(x*x)
          xnum = p(6)*xsq
          xden = xsq
          xnum = (xnum+p(1))*xsq
          xden = (xden+q(1))*xsq
          xnum = (xnum+p(2))*xsq
          xden = (xden+q(2))*xsq
          xnum = (xnum+p(3))*xsq
          xden = (xden+q(3))*xsq
          xnum = (xnum+p(4))*xsq
          xden = (xden+q(4))*xsq
          result = xsq*(xnum+p(5))/(xden+q(5))
          result = (sqrpi-result)/y
          xsq = AINT(x*sixten)/sixten
          del = (x-xsq)*(x+xsq)
          result = EXP(-xsq*xsq*half)*EXP(-del*half)*result
          ccum = one - result

          IF (x>zero) THEN
            temp = result
            result = ccum
            ccum = temp
          END IF

        END IF

        IF (result<min) result = 0.0E0_dpkind
        IF (ccum<min) ccum = 0.0E0_dpkind

      END SUBROUTINE local_cum_normal

!*********************************************************************

      FUNCTION local_inv_normal(cum,ccum)
!----------------------------------------------------------------------
! ALGORITHM AS241  APPL. STATIST. (1988) VOL. 37, NO. 3
! Produces the normal deviate Z corresponding to a given lower
! tail area of P; Z is accurate to about 1 part in 10**16.
! cum is the desired area to the left of local_inv_normal
! ccum is the desired area to the right
! A higher level of code checks ranges on cum and ccum and assures
! that they sum to one.
! This ELF90-compatible version by Alan Miller - 20 August 1996
! N.B. The original algorithm is as a function; this is a subroutine
! Code modified  (but not in logic) by Barry W. Brown
! May,2001 for incorporation into dcdflib90.
! The code is again a function
!----------------------------------------------------------------------
! .. Implicit None Statement ..
        IMPLICIT NONE
! ..
! .. Parameters ..
        REAL (dpkind), PARAMETER :: a(8) = (/ &
          3.3871328727963666080E0_dpkind, 1.3314166789178437745E+2_dpkind &
          , 1.9715909503065514427E+3_dpkind, &
          1.3731693765509461125E+4_dpkind, &
          4.5921953931549871457E+4_dpkind, &
          6.7265770927008700853E+4_dpkind, &
          3.3430575583588128105E+4_dpkind, &
          2.5090809287301226727E+3_dpkind/)
        REAL (dpkind), PARAMETER :: b(8) = (/ 1.0_dpkind, &
          4.2313330701600911252E+1_dpkind, &
          6.8718700749205790830E+2_dpkind, &
          5.3941960214247511077E+3_dpkind, &
          2.1213794301586595867E+4_dpkind, &
          3.9307895800092710610E+4_dpkind, &
          2.8729085735721942674E+4_dpkind, &
          5.2264952788528545610E+3_dpkind/)
        REAL (dpkind), PARAMETER :: c(8) = (/ &
          1.42343711074968357734E0_dpkind, &
          4.63033784615654529590E0_dpkind, &
          5.76949722146069140550E0_dpkind, &
          3.64784832476320460504E0_dpkind, &
          1.27045825245236838258E0_dpkind, &
          2.41780725177450611770E-1_dpkind, &
          2.27238449892691845833E-2_dpkind, &
          7.74545014278341407640E-4_dpkind/)
        REAL (dpkind), PARAMETER :: d(8) = (/ 1.0_dpkind, &
          2.05319162663775882187E0_dpkind, &
          1.67638483018380384940E0_dpkind, &
          6.89767334985100004550E-1_dpkind, &
          1.48103976427480074590E-1_dpkind, &
          1.51986665636164571966E-2_dpkind, &
          5.47593808499534494600E-4_dpkind, &
          1.05075007164441684324E-9_dpkind/)
        REAL (dpkind), PARAMETER :: e(8) = (/ &
          6.65790464350110377720E0_dpkind, &
          5.46378491116411436990E0_dpkind, &
          1.78482653991729133580E0_dpkind, &
          2.96560571828504891230E-1_dpkind, &
          2.65321895265761230930E-2_dpkind, &
          1.24266094738807843860E-3_dpkind, &
          2.71155556874348757815E-5_dpkind, &
          2.01033439929228813265E-7_dpkind/)
        REAL (dpkind), PARAMETER :: f(8) = (/ 1.0_dpkind, &
          5.99832206555887937690E-1_dpkind, &
          1.36929880922735805310E-1_dpkind, &
          1.48753612908506148525E-2_dpkind, &
          7.86869131145613259100E-4_dpkind, &
          1.84631831751005468180E-5_dpkind, &
          1.42151175831644588870E-7_dpkind, &
          2.04426310338993978564E-15_dpkind/)
        REAL (dpkind), PARAMETER :: const1 = 0.180625E0_dpkind
        REAL (dpkind), PARAMETER :: const2 = 1.6E0_dpkind
        REAL (dpkind), PARAMETER :: split1 = 0.425E0_dpkind
        REAL (dpkind), PARAMETER :: split2 = 5.E0_dpkind
! ..
! .. Scalar Arguments ..
        REAL (dpkind), INTENT (IN) :: ccum, cum
! ..
! .. Local Scalars ..
        REAL (dpkind) :: p, q, r
! ..
! .. Function Return Value ..
        REAL (dpkind) :: local_inv_normal
! ..
! .. Intrinsic Functions ..
        INTRINSIC ABS, LOG, MIN, SQRT
! ..
! .. Executable Statements ..

        p = MIN(cum,ccum)

        q = p - half

        IF (ABS(q)<=split1) THEN
          r = const1 - q*q
          local_inv_normal = q*evaluate_polynomial(a,r)/ &
            evaluate_polynomial(b,r)

          IF (ccum<cum) local_inv_normal = -local_inv_normal
        ELSE
          IF (q<zero) THEN
            r = p
          ELSE
            r = one - p
          END IF

          r = SQRT(-LOG(r))

          IF (r<=split2) THEN
            r = r - const2
            local_inv_normal = evaluate_polynomial(c,r)/ &
              evaluate_polynomial(d,r)
          ELSE
            r = r - split2
            local_inv_normal = evaluate_polynomial(e,r)/ &
              evaluate_polynomial(f,r)
          END IF

          IF ((q<zero) .OR. (ccum<cum)) local_inv_normal = - &
            local_inv_normal

        END IF

      END FUNCTION local_inv_normal

!*********************************************************************

    END MODULE cdf_normal_mod
